from __future__ import annotations
from typing import TYPE_CHECKING, NamedTuple

from smufolib import stdUtils

if TYPE_CHECKING:
    from fontParts.fontshell import RComponent
    from fontParts.fontshell import RContour
    from smufolib.objects.glyph import Glyph


# pylint: disable=invalid-name, too-many-arguments


class Point(NamedTuple):
    """Named tuple for point values."""
    type: str
    position: Position
    contourIndex: int


class Position(NamedTuple):
    """Named tuple for position values."""
    x: int | float
    y: int | float


def getPoints(glyph: Glyph,
              types: str | tuple[str] = ('line', 'curve', 'qcurve')
              ) -> tuple[NamedTuple]:
    """Get tuple of points for a glyph."""
    points = tuple(getContourPoints(c, types) for c in glyph)
    if glyph.components:
        points += tuple(getCompositePoints(c, types) for c in glyph.components)
    return points


def getContourPoints(contour: RContour,
                     types: str | tuple[str]
                     ) -> tuple[NamedTuple]:
    """Get tuple of points for all glyph contours."""
    rawPoints = stdUtils.flatten(contour)
    return tuple(Point(p.type, Position(*p.position), p.contour.index)
                 for p in rawPoints if p.type in types)


def getCompositePoints(component: RComponent,
                       types: str | tuple[str]
                       ) -> tuple[str, tuple[int], int]:
    """Get points and offset position from glyph component."""
    baseGlyph = component.font[component.baseGlyph]
    rawPoints = stdUtils.flatten(baseGlyph.contours)
    offset, index = component.offset, component.index
    return tuple(Point(
        p.type, Position(*stdUtils.addTuples(p.position, offset)),
        index) for p in rawPoints if p.type in types)



# pylint: disable=C0114
from __future__ import annotations
from typing import Any
import contextlib
from pathlib import Path
import json
import urllib.request
import warnings

from smufolib import config
from smufolib.normalizers import normalizeRequestPath

CONFIG = config.load()


class Request:
    """Send HTTP or filesystem request.

    A fallback path (e.g., a filesystem path to the same file), may be
    specified in case of connection failure
    (:class:`urllib.error.URLError`) if the primary path is a URL.

    An optional warning may be raised in the event of a fallback.

    :param path: Primary URL or filepath.
    :param fallback: Fallback filepath to use if path raises URLError.
    :param mode: File usage specification. See :func:`open` for details.
    :param encoding: File text encoding. See :func:`open` for details.
        Defaults to :ref:`[request]` ``encoding`` configuration.
    :param warn: Warn if URLError is raised before fallback request.
        Defaults to :ref:`[request]` ``warn`` configuration.

    """

    # pylint: disable=invalid-name, too-few-public-methods

    def __init__(self, path: Path | str | None = None,
                 fallback: Path | str | None = None,
                 mode: str = 'r',
                 encoding: str = CONFIG['request']['encoding'],
                 warn: bool = CONFIG['request']['warn']) -> None:
        self._path = path
        self._fallback = fallback
        self._mode = mode
        self._encoding = encoding
        self._warn = warn

    def __repr__(self):
        return (f"<{self.__class__.__name__} '{self.path}' "
                f"('{self.fallback}') at {id(self)}>")

    def json(self) -> dict[str, Any]:
        """Parse request as JSON."""
        if self.raw is None:
            return None
        return json.loads(self.raw)

    @property
    def raw(self) -> str | bytes:
        """Make a request and returns raw file contents."""
        if (self.path is None and self.fallback is None):
            raise ValueError(f"Path/fallback must be '{Path.__name__}' "
                             "or 'str' not 'NoneType'.")
        # Employ context manager stack for on- and offline use cases.
        with contextlib.ExitStack() as stack:
            try:
                raw = stack.enter_context(urllib.request.urlopen(self.path))
                return raw.read()
            except (urllib.error.URLError, AttributeError) as exc:
                if not self.fallback:
                    raise urllib.error.URLError(
                        f'Could not connect to url: {self.path}.') from exc

                if self._warn:
                    warnings.warn(
                        f"Could not connect to url: {self.path}.", URLWarning)
                raw = stack.enter_context(
                    open(self.fallback, self.mode, encoding=self.encoding))
                return raw.read()

            except ValueError:
                raw = stack.enter_context(
                    open(self.path, self.mode, encoding=self.encoding))
                return raw.read()

    @property
    def path(self) -> str | None:
        """Primary URL or filepath as string."""
        return normalizeRequestPath(self._path)

    @property
    def fallback(self) -> str | None:
        """Fallback URL or filepath as string."""
        return normalizeRequestPath(self._fallback)

    @property
    def mode(self) -> str:
        """File usage specification."""
        return self._mode

    @property
    def encoding(self) -> str:
        """File text encoding."""
        return self._encoding


class URLWarning(Warning):
    """URL connection failure warning."""



# pylint: disable=C0114
from fontParts.fontshell.font import RFont

from smufolib.objects.layer import Layer
from smufolib.objects.smufl import Smufl


class Font(RFont):
    """Environment implementation of :class:`fontParts.base.BaseFont`.

    To import the :class:`Font` object and instantiate a font::

        >>> from smufolib import Font
        >>> font = Font('path/to/my/font.ufo')

    """

    # pylint: disable=too-few-public-methods

    layerClass = Layer

    @property
    def smufl(self) -> Smufl:
        """Font instance of :class:`~smufolib.objects.smufl.Smufl`."""
        return Smufl(font=self)



#!/usr/bin/env python3
# coding: utf-8

# pylint: disable=C0301

"""This script generates a full featured metadata `JSON
<https://www.json.org/json-en.html>`_ file for SMuFL fonts from
metadata sources and user specifications, including the following
metadata sections:

    * `mandatory (fontName and fontVersion)
      <https://w3c.github.io/smufl/latest/specification/font-specific-metadata.html>`_
    * `size <https://w3c.github.io/smufl/latest/specification/font-specific-metadata.html>`_
    * `engravingDefaults <https://w3c.github.io/smufl/latest/specification/engravingdefaults.html>`_
    * `glyphAdvanceWidths <https://w3c.github.io/smufl/latest/specification/glyphadvancewidths.html>`_
    * `glyphsWithAnchors <https://w3c.github.io/smufl/latest/specification/glyphswithanchors.html>`_
    * `glyphsWithAlternates <https://w3c.github.io/smufl/latest/specification/glyphswithalternates.html>`_
    * `glyphBBoxes <https://w3c.github.io/smufl/latest/specification/glyphbboxes.html>`_
    * `ligatures <https://w3c.github.io/smufl/latest/specification/ligatures.html>`_
    * `sets <https://w3c.github.io/smufl/latest/specification/sets.html>`_
    * `optionalGlyphs <https://w3c.github.io/smufl/latest/specification/optionalglyphs.html>`_

Sections are automatically added depending on the font's scope and
assigned values.

This script requires that SMufoLib be installed within its executive
environment. It may also be imported as a module and contains the
following public functions:

    * :func:`generateMetadata` – The scripts program function.
    * :func:`main` - Command line entry point.

"""
from __future__ import annotations
from typing import TYPE_CHECKING, Any
import argparse
import collections
from pathlib import Path
import json

from tqdm import tqdm

from smufolib import Font, Request, cli, config, stdUtils

CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(CONFIG['metadata.paths']['referenceFont'],
                    CONFIG['metadata.fallbacks']['referenceFont'])
VERBOSE = False

# pylint: disable=C0103


def generateMetadata(font: Font | Path | str,
                     targetPath: str | Path,
                     fontData: str | Path | Request = FONT_DATA,
                     verbose: bool = VERBOSE) -> None:
    """Generate metadata JSON file.

    :param font: Object or path to
     targetPath :class:`~smufolib.objects.font.Font`.
    :param targetPath: Target directory for Metadata JSON file.
    :param verbose: Make output verbose. Defaults to ``False``.

    """
    # Check if targetPath exists.
    if not Path(targetPath).exists():
        raise FileNotFoundError("Target path does not exist.")

    # Convert font path to object.
    font = font if isinstance(font, Font) else Font(font)

    # Define print function to be do-nothing if verbose=False.
    verboseprint = print if verbose else lambda *a, **k: None

    metadata = _compileMetadata(font, fontData, verbose, verboseprint)
    filename = Path(targetPath).resolve() / f'{font.smufl.name}_metadata.json'

    verboseprint("Writing metadata to:", filename)
    with open(filename, 'w', encoding='utf-8') as outfile:
        json.dump(metadata, outfile, indent=4, sort_keys=False)

    print("Done!")


def main():
    """Command line entry point."""
    args = _parseArgs()
    generateMetadata(args.font,
                     args.targetPath,
                     fontData=args.fontData,
                     verbose=args.verbose)


def _compileMetadata(font, fontData, verbose, verboseprint) -> dict[str, Any]:
    # Build SMuFL-specific metadata dictionary.
    # pylint: disable=R0912, R0914

    print("Compiling metadata...")

    metadata = collections.defaultdict(dict)

    keys = ('fontName', 'fontVersion', 'designSize',
            'sizeRange', 'engravingDefaults')

    for key in keys:
        attribute = key[4:].lower() if 'font' in key else key
        value = getattr(font.smufl, attribute)
        if attribute == 'engravingDefaults':
            value = value.items()
        elif value is None:
            continue
        metadata[key] = value

    sets, setglyphs = {}, []
    setsTemplate = _getSetsTemplate(fontData)
    attributes = ('name', 'unicode')
    # Set up progress bar if verbose is True
    font = font if verbose else tqdm(font)
    # skipped = []

    for glyph in font:
        if not glyph or not stdUtils.validateClassAttr(glyph, attributes):
            verboseprint("Skipping unnamed or unencoded glyph:", glyph)
            continue

        s = glyph.smufl
        metadata['glyphAdvanceWidths'][s.name] = s.advanceWidth
        metadata['glyphBBoxes'][s.name] = s.bBox

        if s.alternates:
            baseglyphs = {}
            baseglyphs['alternates'] = s.alternates
            metadata['glyphsWithAlternates'][s.name] = baseglyphs

        if s.anchors:
            metadata['glyphsWithAnchors'][s.name] = s.anchors

        if s.isLigature:
            metadata['ligatures'][s.name] = {}
            metadata['ligatures'][s.name]["codepoint"] = s.codepoint
            if s.componentGlyphs:
                metadata['ligatures'][s.name]['componentGlyphs'] = s.componentGlyphs
            metadata['ligatures'][s.name]["description"] = s.description

        if s.isOptional:
            metadata['optionalGlyphs'][s.name] = {}
            metadata['optionalGlyphs'][s.name]['codepoint'] = s.codepoint
            metadata['optionalGlyphs'][s.name]['description'] = s.description
            metadata['optionalGlyphs'][s.name]['classes'] = s.classes

        if s.isSet:
            sets[s.suffix] = setsTemplate[s.suffix]
            glyphinfo = {"alternateFor": s.base.name,
                         "codepoint": s.codepoint,
                         "description": s.description,
                         "name": s.name}
            setglyphs.append(glyphinfo)
            metadata['sets'] = sets
            metadata['sets'][s.suffix]['glyphs'] = setglyphs

    return metadata


def _getSetsTemplate(fontData) -> dict[str, dict[str, str]]:
    # Extract sets template from bravura_metadata.json.
    sets = {}
    metadata = fontData.json()
    for key, values in metadata['sets'].items():
        sets[key] = {}
        for subKey, value in values.items():
            if subKey not in ('description', 'type'):
                continue
            sets[key][subKey] = value
    return sets


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser('font',
                              'targetPath',
                              addHelp=True,
                              description="Generate font metadata JSON file.",
                              fontData=FONT_DATA,
                              verbose=VERBOSE)

    return parser.parse_args()


if __name__ == '__main__':
    main()



from __future__ import annotations
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import json
import textwrap

from smufolib import config, converters
from smufolib.objects.font import Font
from smufolib.request import Request

CONFIG = config.load()

# pylint: disable=invalid-name, unused-argument, too-many-branches


def commonParser(*args, addHelp: bool = False, description: str | None = None,
                 **kwargs) -> ArgumentParser:
    r"""Provide generic command-line arguments and options.

    See the :ref:`Available Options` for details.

    :param \*args: Required positional arguments to assign.
    :param addHelp: Add help message. Should be ``False`` when
        function is parent and otherwise ``True``.
    :param description: Program description when used directly.
    :param \**kwargs: Options and their default values to assign.
    :raises TypeError: Duplicate arguments in \*args and \**kwargs.

    Examples::

        >>> import argparse
        >>> from smufolib import Font, cli
        >>> args = cli.commonParser('font', clear=True)
        >>> parser = argparse.ArgumentParser(parents=[args],
        ...             description='showcase commonParser')
        >>> parser.add_argument(
        ...     '-O', '--include-optionals',
        ...     action='store_true',
        ...     default=includeOptionals,
        ...     help="include optional glyphs",
        ...     dest='includeOptionals'
        ... )
        >>> parser.parse_args("-h".split()))
        usage: test.py [-h] [-x] [-O] font

        showcase commonParser

        positional arguments:
            font         path to UFO file

        optional arguments:
            -h, --help                show this help message and exit
            -x, --clear               erase preexisting objects on execution
            -O, --include-optionals   include optional glyphs

    ::

        >>> parser.parse_args("-f path/to/my/font.ufo".split())
        Namespace(font=<Font 'MyFont' path='path/to/my/font.ufo' at 4377107232>, mark=False)

    ::

        >>> parser.parse_args("-f path/to/my/font.ufo --clear".split())
        Namespace(font=<Font 'MyFont' path='path/to/my/font.ufo' at 4377107232>, mark=True)

    """
    parser = ArgumentParser(add_help=addHelp,
                            formatter_class=ArgumentDefaultsHelpFormatter,
                            description=description)

    settings = {
        'attributes': {
            'nargs': '+',
            'help': "ID attributes to be set: name, classes and/or description"
        },
        'classesData': {
            'type': Request,
            'help': "path to classes metadata file"
        },
        'clear': {
            'action': 'store_true',
            'help': "erase preexisting objects on execution"
        },
        'color': {
            'nargs': 4,
            'type': converters.toNumber,
            'help': "list of RGBA color values"
        },
        'colors': {
            'type': json.loads,
            'help': textwrap.dedent("""\
            Keys mapped to RGBA color arrays as JSON string, e.g.
            {"mark1": [1, 0, 0, 1], "cutOutNE": [0, 0.8, 0, 1]}
            """)
        },
        'exclude': {
            'nargs': '+',
            'help': "objects to exclude from processing"
        },
        'font': {
            'type': Font,
            'help': "path to UFO file"
        },
        'fontData': {
            'type': Request,
            'help': "path to font metadata file"
        },
        'glyphnamesData': {
            'type': Request,
            'help': "path to glyphnames metadata file"
        },
        'include': {
            'nargs': '+',
            'help': "objects to include in processing"
        },
        'includeOptionals': {
            'action': 'store_true',
            'help': "include optional glyphs"
        },
        'mark': {
            'action': 'store_true',
            'help': "apply defined color values to objects"
        },
        'overwrite': {
            'action': 'store_true',
            'help': "overwrite preexisting values"
        },
        'rangesData': {
            'type': Request,
            'help': "path to ranges metadata file"
        },
        'sourcePath': {
            'type': Request,
            'help': "path to source file or directory"
        },
        'spaces': {
            'action': 'store_true',
            'help': "values are given in staff spaces"
        },
        'targetPath': {
            'help': "path to target file or directory"
        },
        'verbose': {
            'action': 'store_true',
            'help': "make output verbose"
        }
    }

    for arg in args:
        settings[arg]['metavar'] = converters.toKebab(arg)
        parser.add_argument(arg, **settings[arg])

    seen = set()
    for key, value in kwargs.items():
        if key in args:
            raise TypeError(
                f"Option '{key}' is already added as positional argument.")

        flags = _generateFlags(key)
        assert flags[0] not in seen, f"Short flag '{flags[0]}' is duplicate."
        seen.add(flags[0])

        settings[key]['dest'] = key
        if value is not None:
            settings[key]['default'] = value
        parser.add_argument(*flags, **settings[key])
    return parser


def _generateFlags(argument):
    # Generates tuple of option flags.
    shortFlags = CONFIG['cli.shortFlags']
    longFlag = f'--{converters.toKebab(argument)}'
    return (shortFlags[argument], longFlag)



#!/usr/bin/env python3
# coding: utf-8

"""This script imports :class:`~smufolib.objects.smufl.Smufl`
identification attribute values for all the glyphs in a SMufoLib
:class:`~smufolib.objects.font.Font` from metadata files. More
specifically, it automatically sets the values for
:attr:`~smufolib.objects.smufl.Smufl.name`,
:attr:`~smufolib.objects.smufl.Smufl.description` and/or
:attr:`~smufolib.objects.smufl.Smufl.classes`, based on the official metadata
files of `SMuFL <https://w3c.github.io/smufl/latest/specification/smufl-metadata.html>`_
and `Bravura <https://github.com/steinbergmedia/bravura#bravura-music-font>`_,
or any other compatible sources.

If ``includeOptionals=True``, optional alternates and ligatures
require that such glyphs be named with reference to their base glyph
(see :ref:`this note <about glyph naming>` for more information about
glyph naming).

If ``overwrite=False``, glyphs with preset attributes are skipped.
Glyphs are also skipped if they are non-SMuFL members or if lookup
fails (because the glyph is unencoded or the codepoint does not exist
in the source metadata.

This script requires that SMufoLib be installed within its
executive environment. It may also be imported as a module and contains
the following public functions:

    * :func:`importID` – The scripts program function.
    * :func:`main` - Command line entry point.

"""
from __future__ import annotations
import argparse
import collections
from pathlib import Path

from tqdm import tqdm

from smufolib import Font, Request, cli, config, converters

CONFIG = config.load()

# Parameter defaults
CLASSES_DATA = Request(CONFIG['metadata.paths']['classes'],
                       CONFIG['metadata.fallbacks']['classes'])
GLYPHNAMES_DATA = Request(CONFIG['metadata.paths']['glyphnames'],
                          CONFIG['metadata.fallbacks']['glyphnames'])
FONT_DATA = Request(CONFIG['metadata.paths']['referenceFont'],
                    CONFIG['metadata.fallbacks']['referenceFont'])
ID_ATTRIBUTES = ('name', 'classes', 'description')
INCLUDE_OPTIONALS = False
OVERWRITE = False
VERBOSE = False


# pylint: disable=R0913, R0914, C0103


def importID(font: Font | Path | str,
             attributes: str | tuple[str, ...] = ID_ATTRIBUTES,
             classesData: Request | Path | str = CLASSES_DATA,
             glyphnamesData: Request | Path | str = GLYPHNAMES_DATA,
             fontData: Request | Path | str = FONT_DATA,
             includeOptionals: bool = INCLUDE_OPTIONALS,
             overwrite: bool = OVERWRITE,
             verbose: bool = VERBOSE
             ) -> None:
    """Import SMuFL identification attributes.

    :param font: Object or path to target :class:`~smufolib.Font`.
    :param attributes: ID attributes to be set. Value can be either
     ``'name'``, ``'classes'``, ``'description'`` or :class:`tuple` of
     several. Defaults to all.
    :param classesData: Object call or direct path to classes metadata
     file. Defaults to :class:`~smufolib.request.Request`
     with :attr:`~smufolib.request.Request.path`
     and :attr:`~smufolib.request.Request.fallback` set to
     :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
     ``classes`` configurations.
    :param glyphnamesData: Object call or direct path to glyphnames
     metadata file. Defaults
     to :class:`~smufolib.request.request.Request`
     with :attr:`~smufolib.request.Request.path`
     and :attr:`~smufolib.request.Request.fallback` set to
     :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
     ``glyphnames`` configurations.
    :param fontData: Object call or direct path to reference font
     metadata file. Defaults to :class:`~smufolib.request.Request`
     with :attr:`~smufolib.request.Request.path`
     and :attr:`~smufolib.request.Request.fallback` set to
     :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
     ``referenceFont`` configurations.
    :param includeOptionals: Include optional glyphs. Defaults to
     ``False``.
    :param overwrite: Overwrite preexisting values. Defaults to
     ``False``.
    :param verbose: Make output verbose. Defaults to ``False``.
    :raises TypeError: if any ``attributes`` value is not of the
     accepted type.
    :raises ValueError: if ``attributes`` value is not a valid ID
     attribute.

    """
    print("Processing...", end="\n\n")

    # Convert font path to object.
    font = font if isinstance(font, Font) else Font(font)

    # Define print function to be do-nothing if verbose=False.
    verboseprint = print if verbose else lambda *a, **k: None

    attributes = _normalizeAttributes(attributes)

    try:
        classesDataJson = classesData.json()
        glyphnamesDataJson = glyphnamesData.json()
        fontDataJson = fontData.json()
    except AttributeError:
        classesDataJson = Request(classesData).json()
        glyphnamesDataJson = Request(glyphnamesData).json()
        fontDataJson = Request(fontData).json()

    print("Building glyphmaps...")
    glyphMaps = _buildGlyphMaps(font, attributes,
                                includeOptionals, classesDataJson,
                                glyphnamesDataJson, fontDataJson)

    print("\n\nSetting attributes...")
    # Use tqdm only when verbose=False.
    for glyph in font if verbose else tqdm(font):
        codepoint = glyph.unicode
        if not codepoint:
            verboseprint("Skipping unencoded glyph:", glyph)
            continue
        if not glyph.smufl.isMember:
            continue
        if glyph.smufl.isOptional and not includeOptionals:
            continue

        for attribute in attributes:
            if codepoint not in glyphMaps[attribute]:
                verboseprint(f"Unable to lookup {attribute} for glyph:", glyph)
                continue

            if (getattr(glyph.smufl, attribute) is not None and not overwrite):
                continue
            setattr(glyph.smufl, attribute, glyphMaps[attribute][codepoint])
            verboseprint(f"Setting {attribute} for glyph:", glyph)

    font.save()
    print('Done!')


def main():
    """Command line entry point."""
    args = _parseArgs()
    importID(args.font,
             args.attributes,
             classesData=args.classesData,
             glyphnamesData=args.glyphnamesData,
             fontData=args.fontData,
             includeOptionals=args.includeOptionals,
             overwrite=args.overwrite,
             verbose=args.verbose)


def _normalizeAttributes(value: str | tuple[str, ...] | list[str]):
    # Normalize values in the ``attributes`` parameter.

    if isinstance(value, str):
        value = (value,)
    try:
        for val in value:
            if val not in ID_ATTRIBUTES:
                raise ValueError("Attribute names must be 'name', 'classes' "
                                 f"or 'description', not '{val}'.")
        return value
    except TypeError as exc:
        raise TypeError("Attributes must be string or tuple of strings, "
                        f"not {type(value.__name__)}") from exc


def _buildGlyphMaps(font, attributes, includeOptionals, classesDataJson,
                    glyphnamesDataJson, fontDataJson) -> dict[str, dict]:
    # Build dictionary of ID attributes mapped to glyphMaps.

    def _buildGlyphMap(attribute, metadata) -> dict[int, str]:
        # Build dictionary of codepoints mapped to annotation values.

        # Descriptions for ligatures are not included in
        # the 'OptionalGlyphs' section of bravura_metadata.json, and
        # lookups in both sections are therefore needed.

        glyphMap = {}
        if 'optionalGlyphs' in metadata:
            metadata = metadata['optionalGlyphs'] | metadata['ligatures']
        for name, data in metadata.items():
            codepoint = converters.toDecimal(data['codepoint'])
            if attribute == 'name':
                glyphMap[codepoint] = name
            else:
                glyphMap[codepoint] = data.get(attribute, None)
        return glyphMap

    def _buildClassMap(font, classesDataJson, glyphnamesDataJson
                       ) -> dict[int, list[str]]:
        # Build glyphMap for 'classes'.

        # GlyphnamesData has a number of unused codepoints and
        # codepoints for text based fonts (58 as of this writing) which
        # should be skipped.

        # Invert classesData dictionary to glyph names mapped to classes.
        classes = collections.defaultdict(list)
        for clas, glyphs in classesDataJson.items():
            for glyph in glyphs:
                name = converters.toUniName(
                    glyphnamesDataJson[glyph]['codepoint'])
                classes[name].append(clas)

        # Loop through basenames to get classes for optional glyphs and
        # index on codepoints.
        classMap = {}
        for glyph in tqdm(font):
            if not glyph.smufl.codepoint or not glyph.smufl.base:
                continue
            if glyph.smufl.base.name not in classes:
                continue
            if includeOptionals:
                classMap[glyph.unicode] = classes[glyph.smufl.base.name]
            elif not glyph.smufl.isOptional:
                classMap[glyph.unicode] = classes[glyph.smufl.base.name]
        return classMap

    glyphMaps = {}
    for attribute in attributes:
        glyphMaps[attribute] = _buildGlyphMap(attribute, glyphnamesDataJson)
        if includeOptionals:
            glyphMaps[attribute] |= _buildGlyphMap(attribute, fontDataJson)
        if attribute == 'classes':
            glyphMaps[attribute] = _buildClassMap(
                font, classesDataJson, glyphnamesDataJson)
    return glyphMaps


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.

    parser = argparse.ArgumentParser(
        parents=[cli.commonParser(
            'font',
            attributes=ID_ATTRIBUTES,
            classesData=CLASSES_DATA,
            glyphnamesData=GLYPHNAMES_DATA,
            fontData=FONT_DATA,
            includeOptionals=INCLUDE_OPTIONALS,
            overwrite=OVERWRITE,
            verbose=VERBOSE)],
        description='Set annotation attributes from SMuFL metadata.')

    return parser.parse_args()


if __name__ == '__main__':
    main()



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path


from fontParts.base.normalizers import normalizeBoolean  # pylint: disable=unused-import
from fontParts.base.normalizers import normalizeColor  # pylint: disable=unused-import
from fontParts.base.normalizers import normalizeInternalObjectType  # pylint: disable=unused-import
from fontParts.base.normalizers import normalizeVisualRounding  # pylint: disable=unused-import
from fontParts.base.normalizers import normalizeGlyph  # pylint: disable=unused-import

if TYPE_CHECKING:
    from smufolib.objects.font import Font
    from smufolib.request import Request
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.engravingDefaults import EngravingDefaults

# pylint: disable=C0103, C0415


# ----
# Font
# ----

def normalizeFont(value: Font) -> Font:
    """Normalize Font object.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not the accepted type.

    """
    from smufolib.objects.font import Font
    return normalizeInternalObjectType(value, Font, "Font object")


# -----
# Smufl
# -----

def normalizeClasses(value: tuple[str, ...] | list[str] | None
                     ) -> tuple[str] | None:
    """Normalize smufl classes.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not an accepted type.
    :raises ValueError:
        - if any ``value`` item does not normalize.
          with :func:`normalizeSmuflName`.
        - if any ``value`` item is :obj:`None`.
        - if any ``value`` items are duplicates.

    """
    if value is None:
        return None
    if not isinstance(value, (tuple, list)):
        raise TypeError(
            f"Classes must be tuple, list or None, not {type(value).__name__}.")

    for val in value:
        if val is None:
            raise ValueError("Class names cannot be None.")
        for v in val.split('_'):
            normalizeSmuflName(v)

    duplicates = {v for v in value if value.count(v) > 1}
    if len(duplicates) != 0:
        raise ValueError(
            "Duplicate class names are not allowed. Class "
            f"name(s) '{','.join(duplicates)}' are duplicate.")
    return tuple(v for v in value if v)


def normalizeDescription(value: str | None) -> str | None:
    """Normalize smufl descriptions.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not an accepted type.
    :raises ValueError: if ``value`` is an empty string.

    """
    if value is None:
        return None
    try:
        if len(value) == 0:
            raise ValueError("The description is empty.")
        return value
    except TypeError as exc:
        raise TypeError("Descriptions must be strings or None, "
                        f"not {type(value).__name__}.") from exc


def normalizeDesignSize(value: int | None) -> int | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not an accepted type.
    :raises ValueError: if ``value`` is less than 10.

    """
    if value is None:
        return None
    try:
        if value < 10:
            raise ValueError("Design size must not be less than 10.")
        return value
    except TypeError as exc:
        raise TypeError("Design size must be an integer or None, "
                        f"not {type(value).__name__}.") from exc


def normalizeSizeRange(value: tuple[int, int] | list[int] | None
                       ) -> tuple[int, int] | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not an accepted type.
    :raises ValueError:
        - if ``value`` does not contain a value pair.
        - if ``value`` is not an increasing range.
        - if ``value`` items do not normalize
          with :func:`normalizeDesignSize`, except not be :obj:`None`.

    """
    if value is None:
        return None
    try:
        length = len(value)
        if length != 2:
            raise ValueError(f"Size range must be of length 2, not {length}.")
        start, end = value
        if start >= end:
            raise ValueError("Size range values must be an increasing range.")
        return tuple(normalizeDesignSize(start), normalizeDesignSize(end))
    except TypeError as exc:
        raise TypeError(f"Size range must be a list or None, "
                        f"not {type(value).__name__}.") from exc


def normalizeSmufl(value: Smufl) -> Smufl:
    """Normalize Smufl object.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not the accepted type.

    """
    from smufolib.objects.smufl import Smufl
    return normalizeInternalObjectType(value, Smufl, "Smufl object")


def normalizeSmuflName(value: str | None) -> str | None:
    """Normalize smufl names.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not an accepted type.
    :raises ValueError:
        - if ``value`` is an empty string.
        - if ``value`` is value contains non-alphanumeric characters.
        - if ``value`` does not start with a lowercase letter or number.

    """
    if value is None:
        return None
    try:
        if len(value) == 0:
            raise ValueError("The name is empty.")
        for val in value:
            if not val.isalnum():
                raise ValueError(f"The name '{value}' contains "
                                 f"an invalid character: '{val}'.")
        if value[0].isupper():
            raise ValueError(
                "Names must start with a lowercase letter or number.")
        return value
    except TypeError as exc:
        raise TypeError("Names must be str or None, "
                        f"not {type(value).__name__}.") from exc


# ------------------
# Engraving defaults
# ------------------

def normalizeEngravingDefaults(value: EngravingDefaults) -> EngravingDefaults:
    """Normalize EngravingDefaults object.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not the accepted type.

    """
    from smufolib.objects.engravingDefaults import EngravingDefaults
    return normalizeInternalObjectType(
        value, EngravingDefaults, "EngravingDefaults object")


def normalizeEngravingDefaultsAttr(name: str,
                                   value: int | float
                                   | tuple[str, ...] | list[str] | None
                                   ) -> int | float | tuple[str, ...] | None:
    """Normalize engraving defaults attribute value based on name.

    :param name: The name of the attribute to normalize.
    :param value: The value to normalize.
    :raises TypeError: if ``name`` or ``value`` is not an accepted
     type.
    :raises ValueError: if ``name='textFontFamily'`` and ``value``
     contains an empty string.

    """
    if value is None:
        return None
    if name == 'textFontFamily':
        value = _normalizeStringList(name, value)
    elif not isinstance(value, (int, float)):
        raise TypeError(f"{name} must be an int, float or None, "
                        f"not {type(value).__name__}.")
    return value


def _normalizeStringList(name: str,
                         value: tuple['str', ...] | list[str]
                         ) -> tuple['str']:
    # Normalize string list.
    if not isinstance(value, (tuple, list)):
        raise TypeError(f"{name} must be a list or None, "
                        f"not {type(value).__name__}.")
    for val in value:
        if not isinstance(val, str):
            raise TypeError(f"{name} items must be strings, "
                            f"not {type(value).__name__}.")
        if len(val) == 0:
            index = value.index(val)
            raise ValueError(f"{name} item at index {index} is empty.")
    return tuple(value)


# -------
# Request
# -------


def normalizeRequest(value: Request) -> Request:
    """Normalize Request object.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not the accepted type.

    """
    from smufolib.request import Request
    return normalizeInternalObjectType(value, Request, "Request object")


def normalizeRequestPath(value: str | Path | None) -> str | None:
    """Normalize Request path.

    Relative paths are resolved automatically.

    :param value: The value to normalize.
    :raises TypeError: if ``value`` is not the accepted type.

    """
    if value is None:
        return None
    try:
        if isinstance(value, Path) or value.startswith('.'):
            return str(Path(value).resolve())
        return value
    except TypeError as exc:
        raise TypeError("Expected str or Path, not "
                        f"'{value.__class__.__name__}'.") from exc



"""Test converters module for SMufoLib."""

import unittest
from smufolib import converters

# pylint: disable=missing-function-docstring, invalid-name


class Converters(unittest.TestCase):
    def test_convertMeasurement(self):
        self.assertEqual(converters.convertMeasurement(
            125, convertTo='spaces', unitsPerEm=1000), 0.5)
        self.assertEqual(converters.convertMeasurement(
            0.5, convertTo='units', unitsPerEm=1000), 125)
        with self.assertRaises(TypeError):
            converters.convertMeasurement(
                (125,), convertTo='spaces', unitsPerEm=1000)
        with self.assertRaises(ValueError):
            converters.convertMeasurement(
                125, convertTo='something else', unitsPerEm=1000)

    def test_toDecimal(self):
        for value in ('U+E00C', 'uE00C', 'uniE00C'):
            self.assertEqual(converters.toDecimal(value), 57356)
            with self.assertRaises(ValueError):
                converters.toDecimal(value[1:])
        with self.assertRaises(TypeError):
            converters.toDecimal(57356)


# def testToUniHex():
#     for value in (57344, '57344'):
#         assert converters.toUniHex(value) == 'U+E000'

#     with pytest.raises(ValueError):
#         converters.toUniHex('23a45')


# def testToUniName():
#     for value in (57344, '57344', 'U+E000'):
#         assert converters.toUniName(value, short=False) == 'uniE000'
#         assert converters.toUniName(value, short=True) == 'uE000'

#     with pytest.raises(ValueError):
#         for value in ('U+ E000', 'U+E0G0'):
#             converters.toUniName(value)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# coding: utf-8

"""This script checks font anchors against the metadata file of SMuFL's
reference font, Bravura, to find missing or superfluous glyph anchors
according to the SMuFL standard.

Discrepancies are printed to console, and glyphs may be marked with
color value specified in smufolib.cfg. Glyphs without annotated smufl
names or glyph names will be skipped.

This script requires that SMufoLib be installed within its executive
environment. It may also be imported as a module and contains the
following public funcitons:

    * :func:`checkAnchors` - The scripts program function.
    * :func:`main` – Command line entry point.

"""
from __future__ import annotations
from typing import Any
import argparse
from pathlib import Path

from tqdm import tqdm

from smufolib import Font, Request, cli, config, normalizers

CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(CONFIG['metadata.paths']['referenceFont'],
                    CONFIG['metadata.fallbacks']['referenceFont'])
MARK = False
MARK_COLOR = CONFIG['color.marks']['mark1']
VERBOSE = False

# pylint: disable=invalid-name


def checkAnchors(font: Font | Path | str,
                 fontData: Request | Path | str = FONT_DATA,
                 mark: bool = MARK,
                 color: tuple[int | float] | None = MARK_COLOR,
                 verbose: bool = VERBOSE) -> None:
    """Check validity of SMuFL-specific glyph anchors.

    :param font: Object or path to
     target :class:`~smufolib.objects.font.Font`.
    :param fontData: Object call or direct path to reference font
     metadata file. Defaults to :class:`~smufolib.request.Request`
     with :attr:`~smufolib.request.Request.path`
     and :attr:`~smufolib.request.Request.fallback` set to
     :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
     ``referenceFont`` configurations.
    :param mark: mark discrepant glyphs. Defaults to ``False``.
    :param color: Color value to apply when ``mark=True``.
     Defaults to
     :ref:`[color.marks]` ``mark1`` configuration.
     :param verbose: Make output verbose. Defaults to ``False``.

    """
    names = {}
    fontAnchors = {}
    referenceAnchors = {}
    if color is not None:
        color = normalizers.normalizeColor(color)

    # Convert font path to object.
    font = font if isinstance(font, Font) else Font(font)

    # Define print function to be do-nothing if verbose=False.
    verboseprint = print if verbose else lambda *a, **k: None

    # Build dicts of glyph names and anchors indexed on smufl names.
    print("Processing...")
    for glyph in tqdm(font):
        names[glyph.smufl.name] = glyph.name
        if not glyph.anchors:
            continue
        fontAnchors[glyph.smufl.name] = [a.name for a in glyph.anchors]

    # Get JSON from metadata path.
    try:
        metadata = fontData.json()
    except AttributeError:
        metadata = Request(fontData).json()

    # Build dict of reference anchors indexed on smufl names.
    for name, anchors in metadata['glyphsWithAnchors'].items():
        if name in fontAnchors:
            referenceAnchors[name] = anchors.keys()

    # Compare reference anchors to font anchors to get deficit.
    if not fontAnchors:
        verboseprint("\nNo anchors found.")
    else:
        verboseprint("\nMissing anchors:".upper(), end='\n\n')
        result = _evaluate(referenceAnchors, fontAnchors, names, verboseprint)
        if mark and color and result:
            for name in result:
                glyph = font[names[name]]
                glyph.markColor = color

        # Compare font anchors to reference anchors to get surplus.
        verboseprint("\nSuperfluous anchors:".upper(), end='\n\n')
        result = _evaluate(fontAnchors, referenceAnchors, names, verboseprint)
        if mark and color and result:
            for name in result:
                glyph = font[names[name]]
                glyph.markColor = color

        font.save()
        print("Done!")


def main():
    """Command line entry point."""
    args = _parseArgs()
    checkAnchors(args.font,
                 fontData=args.fontData,
                 mark=args.mark,
                 color=args.color,
                 verbose=args.verbose)


def _evaluate(test: dict[str:Any],
              reference: dict[str:Any],
              names: dict[str:str],
              verboseprint) -> list[str]:
    # Compare sources and print results.
    findings = []
    if not findings:
        verboseprint("None")
        return None

    for name, anchors in test.items():
        if name not in reference:
            continue

        for anchor in anchors:
            if anchor in reference[name]:
                continue

            if name not in findings:
                findings.append(name)
                verboseprint(f"\t{names[name]} ({name}):")
            verboseprint(f"\t\t{anchor}")

    return findings


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        'font',
        addHelp=True,
        description=(
            "Find missing or superfluous SMuFL anchors."),
        fontData=FONT_DATA,
        mark=MARK,
        color=MARK_COLOR,
        verbose=VERBOSE)
    return parser.parse_args()


if __name__ == '__main__':
    main()



#!/usr/bin python3
# coding: utf-8

"""This script measures and sets engraving defaults based on the
dimensions, registration, metrics and/or point placement of appropriate
font glyphs.

With limited means of identification and control over the necessary
parameters, the script presently relies on a rather rudimentary and
somewhat error-prone process for calculating values. For this reason,
the results should be used only as a starting point for setting the
values of
:class:`~smufolib.objects.engravingDefaults.EngravingDefaults`
attributes.

The user has the option to override or skip any automatic measurements,
as well as reassign attributes to different ruler functions or glyphs
than the default. This is particularly useful for automatic
measurements of line widths for e.g. octave lines, pedal lines or
repeat endings, which are not represented by a particular SMuFL glyph,
but are often based on the thickness of other lines, such as staff
lines, leger lines, barlines, etc., depending on the desired
thickness.

.. note:: The setting
   for
   :attr:`~smufolib.objects.engravingDefaults.EngravingDefaults.textFontFamily`,
   which involves no measuring or calculation, has to be set manually
   with the ``override`` parameter.

This script requires that SMufoLib be installed within its executive
environment. It may also be imported as a module and contains the
following public functions:

    * :func:`calculateEngravingDefaults` - The scripts program function.
    * :func:`main` – Command line entry point.
    * :func:`boundsLeft` – Returns absolute value of bounds x minimum.
    * :func:`boundsHeight` – Returns absolute value of bounds height.
    * :func:`stemDot` – Measures distance between stem and dot
       countour.
    * :func:`xInner` – Measures distance between two adjacent x points
       of different contours.
    * :func:`xOrigin` – Measures distance between two adjacent x points
      closest to origin.
    * :func:`yInner` – Measures distance between two adjacent y points
       of different contours.
    * :func:`yMinimum` – Measures distance between two adjacent
       low-points on y axis.

"""
from __future__ import annotations
from typing import TYPE_CHECKING
import argparse
import json
import textwrap
from pathlib import Path

from smufolib import Font, cli, config, pointUtils

if TYPE_CHECKING:
    from fontParts.fontshell.point import RPoint
    from smufolib.objects.glyph import Glyph

CONFIG = config.load()

# Parameter defaults
EXCLUDE = None
OVERRIDE = None
SPACES = False
REMAP = None
MARGIN_OF_ERROR = 6

# pylint: disable=invalid-name, too-many-arguments


def calculateEngravingDefaults(font: Font | Path | str,
                               exclude: str | list | None = EXCLUDE,
                               override: dict[str, int |
                                              float] | None = OVERRIDE,
                               remap: dict[str, dict[str, str | int]
                                           ] | None = REMAP,
                               spaces: bool = SPACES) -> None:
    """Calculate engraving defaults from glyph contours.

    :param font: Target font object or path to file.
    :param exclude: :class:`~smufolib.objects.engravingDefaults.EngravingDefaults`
     attributes to exclude. Defaults to ``None``.
    :param override: :class:`~smufolib.objects.engravingDefaults.EngravingDefaults`
     attributes to manually override mapped to their values. Defaults to ``None``.
    :param remap: :class:`~smufolib.objects.engravingDefaults.EngravingDefaults`
     attributes mapped to remappings :class:`dict`, e.g.:

        .. code-block:: python

            r = {
                'arrowShaftThickness': {
                    'ruler': 'boundsLeft',
                    'glyph': 'uniEB60',
                    'referenceIndex': 2
                }
            }

     Defaults to ``None``.
    :param spaces: set units of measurement to font units
     or :attr:`~smufolib.objects.smufl.Smufl.spaces`. Defaults to
     ``False``.

    """
    dispatcher = {
        'boundsHeight': boundsHeight,
        'boundsLeft': boundsLeft,
        'stemDot': stemDot,
        'xInner': xInner,
        'xOrigin': xOrigin,
        'yInner': yInner,
        'yMinimum': yMinimum,
    }

    defaults = {
        'arrowShaftThickness': (xOrigin, 'uniEB60'),
        'barlineSeparation': (xInner, 'uniE031'),
        'beamSpacing': (yInner, 'uniE1F9'),
        'beamThickness': (boundsHeight, 'uniE1F7'),
        'bracketThickness': (xOrigin, 'uniE003'),
        'dashedBarlineDashLength': (yMinimum, 'uniE036'),
        'dashedBarlineGapLength': (yInner, 'uniE036'),
        'dashedBarlineThickness': (xOrigin, 'uniE036'),
        'hairpinThickness': (yMinimum, 'uniE53E'),
        'hBarThickness': (yMinimum, 'uniE4F0'),
        'legerLineExtension': (boundsLeft, 'uniE022'),
        'legerLineThickness': (boundsHeight, 'uniE022'),
        'lyricLineThickness': (boundsHeight, 'uniE010'),
        'octaveLineThickness': (boundsHeight, 'uniE010'),
        'pedalLineThickness': (boundsHeight, 'uniE010'),
        'repeatBarlineDotSeparation': (stemDot, 'uniE040'),
        'repeatEndingLineThickness': (xOrigin, 'uniE030'),
        'slurEndpointThickness': (xOrigin, 'uniE1FD'),
        'slurMidpointThickness': (yMinimum, 'uniE1FD'),
        'staffLineThickness': (boundsHeight, 'uniE010'),
        'stemThickness': (xOrigin, 'uniE210'),
        'subBracketThickness': (xOrigin, 'uniE030'),
        'textEnclosureThickness': (boundsHeight, 'uniE010'),
        'thickBarlineThickness': (xOrigin, 'uniE034'),
        'thinBarlineThickness': (xOrigin, 'uniE030'),
        'thinThickBarlineSeparation': (xInner, 'uniE032'),
        'tieEndpointThickness': (xOrigin, 'uniE1FD'),
        'tieMidpointThickness': (yMinimum, 'uniE1FD'),
        'tupletBracketThickness': (xOrigin, 'uniE1FE')
    }

    font.smufl.spaces = False

    for key, mapping in defaults.items():
        if exclude and key in exclude:
            continue

        ruler, glyph = mapping
        referenceIndex = None
        if remap and key in remap:
            if 'ruler' in remap[key]:
                ruler = dispatcher[remap[key]['ruler']]
            if 'glyph' in remap[key]:
                glyph = remap[key]['glyph']
            referenceIndex = remap[key].get('referenceIndex')

        value = ruler(font[glyph])
        if referenceIndex:
            value = ruler(font[glyph], referenceIndex=referenceIndex)
        if override:
            value = override.get(key)
            if spaces:
                value = font.smufl.toUnits(override.get(key))
            elif 'textFontFamily' in override:
                value = override['textFontFamily']

        setattr(font.smufl.engravingDefaults, key, value)

        if spaces:
            value = font.smufl.toSpaces(value)
        print(f"Setting attribute '{key}': {value}")

    font.save()
    print("Done!")


def main():
    """Command line entry point."""
    args = _parseArgs()
    calculateEngravingDefaults(
        args.font, args.exclude, args.remap, args.override, args.spaces)


# ------
# Rulers
# ------


def boundsHeight(glyph: Glyph) -> int:
    """Height of the glyph bounds.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.

    """
    return glyph.bounds[3] - glyph.bounds[1]


def boundsLeft(glyph: Glyph) -> int:
    """Absolute value of bounds x minimum.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.

    """
    return abs(glyph.bounds[0])


def stemDot(glyph: Glyph, referenceIndex: int=0) -> int:
    """Distance between stem and dot countour.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    curves = sorted(pointUtils.getPoints(glyph, 'curve'),
                    key=lambda p: p.position.x)
    lines = sorted(pointUtils.getPoints(glyph, 'line'),
                   reverse=True, key=lambda p: p.position.x)

    reference = curves[referenceIndex]
    for point in lines:
        if not point.contourIndex != reference.contourIndex:
            continue
        return abs(point.position.x - reference.position.x)


def xInner(glyph: Glyph, referenceIndex: int=3) -> int:
    """Distance between two adjacent x points of different contours.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.x)
    reference = points[referenceIndex]

    for point in points:
        if not (point.contourIndex != reference.contourIndex
                and _areAdjacent(point, reference, axis='y')):
            continue
        return abs(point.position.x - reference.position.x)


def xOrigin(glyph: Glyph, referenceIndex: int=0) -> int:
    """Distance between two adjacent x points closest to origin.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: sum(p.position))
    reference = points[referenceIndex]

    for point in points:
        if not (point.position.x != reference.position.x
                and point.contourIndex == reference.contourIndex
                and _areAdjacent(point, reference, axis='y')):
            continue
        return abs(point.position.x - reference.position.x)


def yOrigin(glyph: Glyph, referenceIndex: int=0) -> int:
    """Distance between two adjacent x points closest to origin.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: sum(p.position))
    reference = points[referenceIndex]

    for point in points:
        if not (point.position.x != reference.position.x
                and point.contourIndex == reference.contourIndex
                and _areAdjacent(point, reference, axis='y')):
            continue
        return abs(point.position.x - reference.position.x)


def yInner(glyph: Glyph, referenceIndex: int=3) -> int:
    """Distance between two adjacent y points of different contours.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.y)
    reference = points[referenceIndex]
    for point in points:
        if not (point.contourIndex != reference.contourIndex
                and _areAdjacent(point, reference, axis='x')):
            continue
        return abs(point.position.y - reference.position.y)


def yMinimum(glyph: Glyph, referenceIndex: int=0) -> int:
    """Distance between two adjacent low-points on axis y.

    :param glyph: Source :class:`~smufolib.objects.glyph.Glyph` of
     contours to measure.
    :param referenceIndex: index of reference point.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.y)
    reference = points[referenceIndex]

    for point in points:
        if not (point.position.y != reference.position.y
                and point.contourIndex == reference.contourIndex
                and _areAdjacent(point, reference, axis='x')):
            continue
        return abs(point.position.y - reference.position.y)


def _areAdjacent(point1: RPoint, point2: RPoint,
                 axis: str | None) -> bool:
    # Check if points are adjacent on axis.
    # Employ margin of error for point placement.

    def withinRange(num1: int, num2: int) -> bool:
        return num2 - MARGIN_OF_ERROR <= num1 <= num2 + MARGIN_OF_ERROR

    x1, x2 = point1.position.x, point2.position.x
    y1, y2 = point1.position.y, point2.position.y

    if axis == 'x':
        checkAxis = withinRange(x1, x2)
    elif axis == 'y':
        checkAxis = withinRange(y1, y2)
    else:
        checkAxis = withinRange(x1, x2) or withinRange(y1, y2)
    return point1 != point2 and checkAxis


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = argparse.ArgumentParser(
        parents=[cli.commonParser('font', exclude=EXCLUDE, spaces=SPACES)],
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__)

    parser.add_argument(
        '-o', '--override',
        default=OVERRIDE,
        type=json.loads,
        help=textwrap.dedent("""\
            JSON string of attributes and values to manually override, e.g.
            {
                "arrowShaftThickness": <number>,
                "barlineSeparation": <number>
            }
            """)
    )

    parser.add_argument(
        '-r', '--remap',
        default=REMAP,
        type=json.loads,
        help=textwrap.dedent("""\
            JSON string of ruler, glyph and referenceIndex remappings, e.g.
            {
                "arrowShaftThickness": {
                    "ruler": "<function name>",
                    "glyph": "<glyph name>",
                    "referenceIndex": <number>
                }
            }
            """)
    )

    return parser.parse_args()


if __name__ == '__main__':
    main()



# pylint: disable=C0114, R0904, C0103, W0212
from __future__ import annotations
from typing import TYPE_CHECKING
import re

from fontParts.base.base import BaseObject

from smufolib.objects.range import Range
from smufolib.objects.engravingDefaults import EngravingDefaults
from smufolib import converters, normalizers
from smufolib.constants import ANCHOR_NAMES

if TYPE_CHECKING:
    from smufolib.objects.layer import Layer
    from smufolib.objects.font import Font
    from smufolib.objects.glyph import Glyph


class Smufl(BaseObject):
    """Provides SMuFL-related metadata and functionality.

    Class may be called from either :class:`~smufolib.objects.font.Font`
    or :class:`~smufolib.objects.glyph.Glyph`. Font-specific attributes
    with unique names may be accessed from both, due to the consistent
    access to parent classes in FontParts.

    .. _about glyph naming:
    .. note:: Attributes with the purpose of identifying ligatures
       (:attr:`isLigature`), stylistic alternates (:attr:`isSalt`) and
       stylistic sets (:attr:`isSet`) depend on strict adherence to the
       descriptive naming schemes stipulated in the `Adobe Glyph List\
       Specification <https://github.com/adobe-type-tools/agl-\
       specification#readme>`_ and followed by the SMuFL standard
       (see `Section 6 <https://github.com/adobe-type-tools/agl-\
       specification#6-assigning-glyph-names-in-new-fonts>`_ for more
       information).

    .. tip:: To avoid having to set all the glyph identification
       attributes manually, it is advisable to run the
       script :mod:`~bin.importID` prior to using this class with
       an excisting font for the first time.

    :param font: Parent :class:`~smufolib.objects.font.Font` object.
    :param glyph: Parent :class:`~smufolib.objects.glyph.Glyph` object.

    While this object is normally created as part of
    a :class:`~smufolib.objects.font.Font`, an orphan :class:`Smufl`
    object can be created like this::

        >>> smufl = Smufl()

    """

    def __init__(self,
                 font: Font | None = None,
                 glyph: Glyph | None = None) -> None:
        self._font = font
        self._glyph = glyph
        self._layer = None

    def _reprContents(self) -> list[str]:
        contents = []
        if self.glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()
        if self.font:
            contents.append("in font")
            contents += self.font._reprContents()
        return contents

    # -------
    # Parents
    # -------

    @property
    def font(self) -> Font | None:
        """Parent :class:`~smufolib.objects.font.Font` object.

        Example::

            >>> glyph.smufl.font
            <Font 'MyFont' path='path/to/my/font.ufo' at 4405856720>

        """
        if self._font is not None:
            return normalizers.normalizeFont(self._font)
        if self._glyph is not None:
            return self._glyph.font
        return None

    @font.setter
    def font(self, value: Font) -> None:
        if self._font is not None and self._font != value:
            raise AssertionError("Font for Smufl object is already "
                                 "set and is not same as value.")
        if self._glyph is not None:
            raise AssertionError("Glyph for Smufl object is already set.")
        self._font = normalizers.normalizeFont(value)

    @property
    def glyph(self) -> Glyph | None:
        """Parent :class:`~smufolib.objects.glyph.Glyph` object.

        Example::

            >>> glyph.smufl.glyph
            <Glyph 'uniE587' ('public.default') at 4536458160>

        """
        if self._glyph is None:
            return None
        return normalizers.normalizeGlyph(self._glyph)

    @glyph.setter
    def glyph(self, value: Glyph) -> None:
        if self._font is not None:
            raise AssertionError("Font for Smufl object is already set.")
        if self._glyph is not None and self._glyph != value:
            raise AssertionError("Glyph for Smufl object is already "
                                 "set and is not same as value.")
        self._glyph = normalizers.normalizeGlyph(value)

    @property
    def layer(self) -> Layer | None:
        """Parent :class:`~smufolib.objects.layer.Layer` object.

        Example::

            >>> glyph.smufl.layer
            <Layer 'public.default' at 4568631952>

        """
        if self._glyph is None:
            return None
        return self._glyph.layer

    # -------------
    # Font Metadata
    # -------------

    @property
    def designSize(self) -> int | None:
        """Optimum point size in integral decipoints.

        Example::

            >>> font.smufl.designSize
            240

        """
        if self.font is None:
            return None
        return self.font.lib.get('_designSize', None)

    @designSize.setter
    def designSize(self, value: int) -> None:
        if value is None:
            self.font.lib.naked().pop('_designSize', None)
        else:
            value = normalizers.normalizeDesignSize(value)
            self.font.lib['_designSize'] = value

    @property
    def engravingDefaults(self) -> EngravingDefaults:
        """Font's :class:`~smufolib.objects.engravingDefaults.EngravingDefaults` object.

        Example::

            >>> font.smufl.engravingDefaults
            <EngravingDefaults at 4540156752>

        """
        return EngravingDefaults(self)

    @property
    def sizeRange(self) -> tuple[int, int] | None:
        """Optimum size range in integral decipoints.

        Example::

            >>> font.smufl.designSize
            (180, 260)

        """
        if self.font is None:
            return None
        return self.font.lib.naked().get('_sizeRange', ())

    @sizeRange.setter
    def sizeRange(self, value: tuple[int, int] | None) -> None:
        if value is None:
            self.font.lib.pop('_sizeRange', None)
        else:
            self.font.lib['_sizeRange'] = normalizers.normalizeSizeRange(value)

    # --------------
    # Glyph metadata
    # --------------

    @property
    def alternates(self) -> tuple[dict[str, str], ...] | None:
        """Metadata of glyph alternates.

        Example::

            >>> glyph = font['uniE050'] # gClef
            >>> glyph.smufl.alternates
            ({'codepoint': 'U+F472', 'name': 'gClefSmall'},)

        """
        if self._glyph is None:
            return None
        # find alt names among string of glyph names
        string = ' '.join(sorted(self.font.keys()))
        pattern = fr'\b{self.glyph.name}\.(?:s?alt|ss)[0-9]{{2}}\b'
        results = re.findall(pattern, string)
        alternates = []
        for name in results:
            glyph = self.font[name]
            alternates.append({'codepoint': glyph.smufl.codepoint,
                               'name': glyph.smufl.name})
        return tuple(alternates)

    @property
    def anchors(self) -> dict[str, tuple[int | float, int | float]] | None:
        """SMuFL-specific glyph anchors as Cartesian coordinates.

        Example::

            >>> glyph = font['uniE0A3'] # noteheadHalf
            >>> glyph.smufl.anchors
            {'cutOutNW': (0.204, 0.296), 'cutOutSE':
            (0.98, -0.3), 'splitStemDownNE': (0.956, -0.3), 'splitStemDownNW':
            (0.128, -0.428), 'splitStemUpSE': (1.108, 0.372), 'splitStemUpSW':
            (0.328, 0.38), 'stemDownNW': (0.0, -0.168), 'stemUpSE': (1.18, 0.168)}

        """

        if self._glyph is None:
            return None
        if self.spaces:
            return {a.name: (self.toSpaces(a.x), self.toSpaces(a.y)) for a in
                    self.glyph.naked().anchors if a.name in ANCHOR_NAMES}
        return {a.name: (a.x, a.y) for a in
                self.glyph.naked().anchors if a.name in ANCHOR_NAMES}

    @property
    def bBox(self) -> dict[str, tuple[int | float, int | float]] | None:
        """Glyph bounding box as Cartesian coordinates.

        Example::

            >>> glyph.smufl.bBox
            {'bBoxSW': (0.0, -0.5), 'bBoxNE': (1.18, 0.5)}

        """
        if self._glyph is None or not self.glyph.bounds:
            return None
        xMin, yMin, xMax, yMax = self.glyph.bounds
        if self.spaces:
            xMin, yMin, xMax, yMax = [
                self.toSpaces(b) for b in self.glyph.bounds]
        return {"bBoxSW": (xMin, yMin), "bBoxNE": (xMax, yMax)}

    @property
    def codepoint(self) -> str | None:
        """Unicode codepoint as formatted string.

        Example::

            >>> glyph = font['uniE0A3']
            >>> glyph.smufl.codepoint
            U+E0A3

        """
        if self._glyph is None or not self.glyph.unicode:
            return None
        return converters.toUniHex(self.glyph.unicode)

    @property
    def componentGlyphs(self) -> tuple[str, ...] | None:
        """Ligature components by :attr:`name`.

        Example::

            >>> glyph = font['uniE09E_uniE083_uniE09F_uniE084']
            >>> glyph.smufl.componentGlyphs
            ('timeSigCombNumerator', 'timeSig3', 'timeSigCombDenominator', 'timeSig4')

        """
        if self._glyph is None:
            return None
        if self.glyph.name is None or '_' not in self.glyph.name:
            return ()
        names = self.glyph.name.split('_')
        components = []
        for name in names:
            if name not in self.font:
                continue
            glyph = self.font[name]
            if not self.name:
                continue
            components.append(glyph.smufl.name)
        return tuple(components)

    @property
    def range(self) -> Range:
        """Glyph's :class:`~smufolib.objects.range.Range` object.

        Example::

            >>> glyph = font['uniE212'] # stemSwished
            >>> glyph.smufl.range
            <Range 'stems' ('U+E210–U+E21F') at 4348391632>

        """
        return Range(self)

    @property
    def advanceWidth(self) -> int | float | None:
        """Glyph advance width.

        Example::

            >>> glyph.smufl.advanceWidth
            671

        """
        if self._glyph is None:
            return None
        if self.spaces:
            return self.toSpaces(self.glyph.width)
        return self.glyph.width

    @advanceWidth.setter
    def advanceWidth(self, value: int | float | None) -> None:
        if self.spaces:
            self.glyph.width = self.toUnits(value)
        else:
            self.glyph.width = value

    # --------------
    # Identification
    # --------------

    # Font
    # ----

    # Font family name is acessible through font.smufl.name.

    @property
    def version(self) -> float | None:
        """SMuFL-specific font version number.

        Example::

            >>> font.smufl.version
            2.2

        """
        if self.font is None:
            return None
        return float(f'{self.font.info.naked().versionMajor}.'
                     f'{self.font.info.naked().versionMinor}')

    @version.setter
    def version(self, value: float | None) -> None:
        if value is None:
            major, minor = None, None
        else:
            major, minor = [int(n) for n in str(value).split('.')]
        self.font.info.naked().versionMajor = major
        self.font.info.naked().versionMinor = minor

    # Glyph
    # -----
    @property
    def classes(self) -> tuple[str, ...] | None:
        """SMuFL-specific class memberships.

        Example::

            >>> glyph = font['uniE354'] # accSagittalSharp7v11kUp
            >>> glyph.smufl.classes
            ['accidentals', 'accidentalsSagittalAthenian', 'combiningStaffPositions']

        """
        if self._glyph is None:
            return None
        return tuple(self.glyph.lib.naked().get('_classes', ()))

    @classes.setter
    def classes(self, value: tuple[str, ...] | list[str] | None) -> None:
        self._set_id('_classes', normalizers.normalizeClasses(value))

    @property
    def description(self) -> str | None:
        """SMuFL-specific glyph description of optional glyphs.

        Example::

            >>> glyph.smufl.description
            Combining swished stem

        """
        if self._glyph is None:
            return None
        return self.glyph.lib.naked().get('_description', None)

    @description.setter
    def description(self, value: str | None) -> None:
        self._set_id('_description', normalizers.normalizeDescription(value))

    @property
    def name(self) -> str | None:
        """SMuFL-specific canonical font or glyph name.

        Examples::

            >>> font.smufl.name
            Bravura

        ::

            >>> glyph = font['uniE212']
            >>> glyph.smufl.name
            stemSwished

        """
        if self._glyph is None:
            if self._font is None:
                return None
            return self.font.info.naked().familyName
        return self.glyph.lib.naked().get('_name', None)

    @name.setter
    def name(self, value: str | None) -> None:
        # Update _names before ID property
        if self._glyph is None:
            self.font.info.naked().familyName = value
        else:
            self._updateNames(normalizers.normalizeSmuflName(value))
            self._set_id('_name', normalizers.normalizeSmuflName(value))

    def _set_id(self, attribute: str, value: str | list | None) -> None:
        # Common ID property setter.
        if self._glyph is None or self._names is None:
            return
        if value is None:
            if attribute not in self.glyph.lib.naked():
                return
            # naked() does not work here.
            self.glyph.lib.pop(attribute, None)
        else:
            self.glyph.lib[attribute] = value

    def _updateNames(self, value) -> None:
        # Keep dynamic dict of names in font.lib.
        if value is None:
            if self._names is None:
                return
            if (len(self._names) == 0 or len(self._names) == 1
                    and self.name in self._names):
                self._names = None
            else:
                self.font.lib.naked()['_names'].pop(self.name, None)
        else:
            if self._names is None:
                self._names = {}
            self._names[value] = self.glyph.name

    @property
    def _names(self) -> dict[str, str] | None:
        # Parent dict in font.lib.naked().
        if self.font is None:
            return None
        return self.font.lib.naked().get('_names')

    @_names.setter
    def _names(self, value: dict[str, str] | None) -> None:
        if value is None:
            self.font.lib.naked().pop('_names', None)
        else:
            self.font.lib.naked()['_names'] = value

    # ----------
    # Validation
    # ----------

    @property
    def isLigature(self) -> bool:
        """Return :obj:`True` if glyph is ligature.

        Example::

            >>> glyph = font['uniE09E_uniE083_uniE09F_uniE084']
            >>> g1.smufl.isLigature
            True
            >>> glyph = font['uniE083']
            >>> glyph.smufl.isLigature
            False

        """
        if (self._glyph is not None and self.glyph.name
            and self.glyph.name.count('uni') > 1
                and '_' in self.glyph.name):
            return True
        return False

    @property
    def isMember(self) -> bool:
        """Return :obj:`True` if glyph is either `recommended or optional
        <https://w3c.github.io/smufl/latest/about/recommended-chars-\
        optional-glyphs.html>`_.

        Example::

            >>> glyph = font['uniE050']
            >>> glyph.smufl.isMember
            True
            >>> glyph = font['spaces']
            >>> glyph.smufl.isMember
            False

        """
        if (self._glyph is not None and self.glyph.unicode
                and 0xe000 <= self.glyph.unicode <= 0xf8ff):
            return True
        return False

    @property
    def isOptional(self) -> bool:
        """Return :obj:`True` if glyph is `optional <https://w3c.github.io\
        /smufl/latest/about/recommended-chars-optional-glyphs.html>`_.

        Example::

            >>> glyph = font['uniF660']
            >>> g1.smufl.isOptional
            True
            >>> glyph = font['uniE083']
            >>> glyph.smufl.isOptional
            False

        """
        if (self._glyph is not None and self.glyph.unicode
                and 0xF400 <= self.glyph.unicode <= 0xF8FF):
            return True
        return False

    @property
    def isRecommended(self) -> bool:
        """Return :obj:`True` if glyph is `recommended <https://w3c.github.io\
        /smufl/latest/about/recommended-chars-optional-glyphs.html>`_.

        Example::

            >>> glyph = font['uniE083']
            >>> g1.smufl.isRecommended
            True
            >>> glyph = font['uniF660']
            >>> glyph.smufl.isRecommended
            False

        """
        if (self._glyph is not None and self.glyph.unicode
                and 0xE000 <= self.glyph.unicode <= 0xF3FF):
            return True
        return False

    @property
    def isSalt(self) -> bool:
        """Return :obj:`True` if glyph is stylistic alternate.

        Accepts both ``'.alt'`` and ``'.salt'`` suffix.
        See :ref:`Note <about glyph naming>` about glyph naming.

        Example::

            >>> glyph = font['uniE042.salt01']
            >>> glyph.smufl.isSalt
            True
            >>> glyph = font['uniE042']
            >>> glyph.smufl.isSalt
            False

        """
        if (self._glyph is not None and self.glyph.name
                and (self.glyph.name.endswith('.salt', 7, -2)
                     or self.glyph.name.endswith('.alt', 7, -2))):
            return True
        return False

    @property
    def isSet(self) -> bool:
        """Return :obj:`True` if glyph is stylistic set member.

        See :ref:`Note <about glyph naming>` about glyph naming.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.isSet
            True
            >>> glyph = font['uniE042.salt01']
            >>> glyph.smufl.isSet
            False

        """
        if (self._glyph is not None and self.glyph.name
                and self.glyph.name.endswith('.ss', 7, -2)):
            return True
        return False

    # -----------------------------
    # Normalization and Measurement
    # -----------------------------

    def round(self) -> None:
        """Round font units to integers.

        Method applies to the following attributes:

        * :attr:`engravingDefaults`
        * :attr:`anchors`
        * :attr:`advanceWidth`

        If :attr:`spaces` is ``True``, values are left unchanged.

        Examples::

            >>> glyph.smufl.advanceWidth
            230.5
            >>> glyph.smufl.round()
            >>> glyph.smufl.wiadvanceWidthdth
            230.5

        ::

            >>> glyph.smufl.spaces = True
            >>> glyph.smufl.advanceWidth
            0.922
            >>> glyph.smufl.round()
            >>> glyph.smufl.advanceWidth
            0.922

        """
        if self.spaces:
            return
        self.engravingDefaults.round()
        if self._glyph is None:
            self.font.round()
        else:
            self.glyph.round()

    def toSpaces(self, value: int | float) -> float | None:
        """Convert font units to staff spaces based on font UPM size.

        The inverse of :meth:`toUnits`.

        :param value: Value to convert.

        Example::

            >>> f = Font('path/to/my/font.ufo')
            >>> font.info.unitsPerEm
            2048
            >>> font.smufl.toSpaces(512)
            1.0

        """
        if self.font is None:
            return None

        return converters.convertMeasurement(
            value,
            convertTo='spaces',
            unitsPerEm=self.font.info.unitsPerEm,
            rounded=False)

    def toUnits(self, value: int | float) -> int | None:
        """Convert staff spaces to font units based on font UPM size.

        The inverse of :meth:`toSpaces`.
        Result is always rounded.

        :param value: Value to convert.

        Example::

            >>> f = Font('path/to/my/font.ufo')
            >>> font.info.unitsPerEm
            2048
            >>> font.smufl.toSpaces(2)
            1024

        """
        if self.font is None:
            return None
        return converters.convertMeasurement(
            value,
            convertTo='units',
            unitsPerEm=self.font.info.unitsPerEm,
            rounded=True)

    @property
    def spaces(self) -> bool:
        """Set state of measurement to staff spaces.

        Example::

            >>> glyph.smufl.advanceWidth
            230.5
            >>> glyph.smufl.spaces = True
            >>> glyph.smufl.advanceWidth
            0.922

        """
        if '_spaces' in self.font.lib.naked():
            return True
        return False

    @spaces.setter
    def spaces(self, value):
        value = normalizers.normalizeBoolean(value)
        if value is False:
            self.font.lib.naked().pop('_spaces', None)
        else:
            self.font.lib.naked()['_spaces'] = True

    # -----
    # Other
    # -----

    @property
    def base(self) -> Glyph | None:
        """:class:`~smufolib.objects.glyph.Glyph` object of alternate's base glyph.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.base
            <Glyph 'uniE050' ('public.default') at 4373577008>

        """
        if self._getBasename():
            return self.font[self._getBasename()]
        return None

    def _getBasename(self) -> str | None:
        # Get name of base glyph.
        if self._glyph is None or self.glyph.name is None:
            return None
        basename = self.glyph.name[:7]
        return basename if basename in self.font else None

    def findGlyph(self, name: str) -> Glyph | None:
        """Find :class:`~smufolib.objects.glyph.Glyph` object from :attr:`name`.

        :param name: SMuFL-specific canonical glyph name.

        Example::

            >>> font.smufl.findGlyph('barlineSingle')
            <Glyph 'uniE030' ('public.default') at 4393557200>

        """
        name = normalizers.normalizeSmuflName(name)
        if self._names is None or name not in self._names:
            return None
        return self.font[self._names[name]]

    @property
    def suffix(self) -> str | None:
        """Return suffix of alternates.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.suffix
            ss01

        """
        if not (self.isSalt or self.isSet):
            return None
        return self.glyph.name.split('.')[1]



#!/usr/bin/env python3
"""This script sets glyph anchors based on data from a SMuFL font metadata
JSON file (SMuFL's reference font, Bravura, by default).

This script requires that SMufoLib be installed within its
executive environment. It may also be imported as a module and contains
the following public funcitons:

    * :func:`importAnchors` – The scripts program function.
    * :func:`main` - Command line entry point.

"""

from __future__ import annotations
from typing import TYPE_CHECKING, Any
import sys
import collections
from pathlib import Path

from smufolib.objects.font import Font
from smufolib.request import Request
from smufolib import cli, config

CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(CONFIG['metadata.paths']['referenceFont'],
                    CONFIG['metadata.fallbacks']['referenceFont'])
MARK = True
COLORS = CONFIG['color.anchors']
CLEAR = False
VERBOSE = False

# pylint: disable=invalid-name, too-many-arguments


def importAnchors(font: Font | Path | str = None,
                  fontData: Request | Path | str = FONT_DATA,
                  mark: bool = MARK,
                  colors: dict[str, tuple[int | float, int | float,
                                          int | float, int | float]] | None = COLORS,
                  clear: bool = CLEAR,
                  verbose: bool = VERBOSE) -> None:
    """Import anchors from font metadata.

    :param font: font object to which the script applies.
    :param fontData: Object call or direct path to reference font
     metadata file. Defaults to :class:`~smufolib.request.Request`
     with :attr:`~smufolib.request.Request.path`
     and :attr:`~smufolib.request.Request.fallback` set to
     :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
     ``referenceFont`` configurations.
    :param mark: Apply defined anchor colors. Defaults to ``True``
    :param colors: dict of anchorNames mapped to UFO color values to
     apply when ``mark=True``.
    :param clear: erase preexisting anchors on append. Defaults to
     ``False``.

    """
    fontData = fontData.json()
    if not isinstance(fontData, Request):
        fontData = Request(fontData).json()

    print('Preparing glyph data...')
    sourceAnchors = fontData['glyphsWithAnchors']
    names = {g.smufl.name: g.name for g in font if g.smufl.name
             and g.smufl.name in sourceAnchors}

    print('Processing...', end='\n\n')
    appended = collections.defaultdict(list)

    for smuflName, anchors in sourceAnchors.items():
        try:
            glyphName = names[smuflName]
        except KeyError:
            continue
        if clear:
            font[glyphName].rGlyph.clearAnchors()
        for anchorName, position in anchors.items():
            if mark and colors:
                anchorColor = colors.get(anchorName, None)
            font[glyphName].appendAnchor(anchorName, position, anchorColor)
            appended[f'{glyphName} ({smuflName})'].append(anchorName)

    font.save()
    if verbose:
        _printDiagnostics(appended)
    print('Done!')


def main():
    """Command line entry point."""
    args = _parseArgs()
    importAnchors(args.font, args.fontData,
                  args.color, args.clear, args.verbose)


def _printDiagnostics(results: dict[str, Any]) -> None:
    print('appended anchors:'.upper())
    for glyphName, anchors in results.items():
        print(f"\n\t{glyphName}:")
        for anchor in anchors:
            print(f"\t\t{anchor}")


def _parseArgs():
    # Parse command line arguments and options.
    parser = cli.commonParser('font',
                              addHelp=True,
                              description=(
                                  "Set anchors from font metadata JSON file."),
                              fontData=FONT_DATA,
                              mark=MARK,
                              colors=COLORS,
                              clear=CLEAR,
                              verbose=VERBOSE)
    return parser.parse_args()


if __name__ == '__main__':
    main()




"""Range metadata module for SMufoLib."""
from __future__ import annotations
from typing import TYPE_CHECKING, List, Dict, Tuple
from pathlib import Path

from smufolib.request import Request
from smufolib import config

if TYPE_CHECKING:
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.glyph import Glyph
    from smufolib.objects.font import Font
    from smufolib.objects.layer import Layer

CONFIG = config.load()
METADATA = Request(CONFIG['metadata.paths']['ranges'],
                   CONFIG['metadata.fallbacks']['ranges']).json()


class Range:
    """SMuFL range-related metadata.

    Information about how a :class:`~smufolib.objects.glyph.Glyph`
    relates to SMuFL's own glyph ranges is accessible through this
    object. It is currently read-only, and retrieves it's data from the
    `ranges.json
    <https://w3c.github.io/smufl/latest/specification/ranges.html>`_
    metadata file.

    :param smufl: Parent :class:`~smufolib.objects.smufl.Smufl`.

    While this object is normally created as part of
    a :class:`~smufolib.objects.font.Font`, an
    orphan :class:`Range` object may be created like
    this::

        >>> range = Range()

    """
    # pylint: disable=invalid-name

    def __init__(self, smufl: Smufl | None = None) -> None:
        self._smufl = smufl

    def __repr__(self):
        return (f"<{self.__class__.__name__} '{self.name}' "
                f"('{self.start}–{self.end}') at {id(self)}>")

    # -------
    # Parents
    # -------

    @property
    def smufl(self) -> Smufl | None:
        """Parent :class:`~smufolib.objects.smufl.Smufl` object."""
        return self._smufl

    @property
    def glyph(self) -> Glyph | None:
        """Parent :class:`~smufolib.objects.glyph.Glyph` object."""
        if self._smufl is None:
            return None
        return self._smufl.glyph

    @property
    def font(self) -> Font | None:
        """Parent :class:`~smufolib.objects.font.Font` object."""
        if self._smufl is None:
            return None
        return self._smufl.font

    @property
    def layer(self) -> Layer | None:
        """Parent :class:`~smufolib.objects.layer.Layer` object."""
        if self._smufl is None:
            return None
        return self._smufl.layer

    # ----
    # Data
    # ----

    @property
    def name(self) -> str | None:
        """Name of affiliated SMuFL range.

        Example::

            >>> range.name
            timeSignatures

        """
        return self._getAttribute('range_name')

    @property
    def description(self) -> str | None:
        """Description of affiliated SMuFL range.

        Example::

            >>> range.description
            Time signatures

        """
        return self._getAttribute('description')

    @property
    def glyphs(self) -> Tuple[Glyph, ...] | None:
        """:class:`~smufolib.objects.glyph.Glyph` objects of Affiliated SMuFL range.

        Example::

            >>> range.glyphs
            (<Glyph 'uniE080' ('public.default') at 4596284000>,
            ...
            <Glyph 'uniE09F' ('public.default') at 4632755792>)

        """
        return self._getAttribute('glyphs')

    @property
    def start(self) -> int | None:
        """Start unicode of affiliated SMuFL range.

        Example::

            >>> range.start
            U+E080

        """
        return self._getAttribute('range_start')

    @property
    def end(self) -> int | None:
        """End unicode of affiliated SMuFL range.

        Example::

            >>> range.end
            U+E09F

        """
        return self._getAttribute('range_end')

    def _getAttribute(self, name: str) -> str | List[str] | None:
        # Get metadata attributes from ranges.json.
        if self.glyph is None or METADATA is None:
            return None
        for range_, attributes in METADATA.items():
            if self.smufl.name not in attributes['glyphs']:
                continue
            if name == 'range_name':
                return range_
            if name == 'glyphs':
                return tuple(self.smufl.findGlyph(n) for n in attributes[name])
            return attributes[name]



"""where SMuFL meets UFO

"""
from smufolib._version import __version__
from smufolib.objects.engravingDefaults import EngravingDefaults
from smufolib.objects.font import Font
from smufolib.objects.glyph import Glyph
from smufolib.objects.layer import Layer
from smufolib.objects.range import Range
from smufolib.objects.smufl import Smufl
from smufolib.request import Request
from smufolib.request import URLWarning
